import json
import math
from typing import List


class Updateable():
    """
    (Base) class that provides update() method for each class that inherits from this class.
    """

    def to_dict(self) -> dict:
        raise NotImplementedError

    def update(self, overlay_object):
        """
        :param overlay_object:  Object that is used to update self.
        """
        # loop over all member variables and try to call update() method, if object does not have update() method just assign new value.
        for v_node, v_overlay_node in zip(vars(self).items(), vars(overlay_object).items()):
            if v_overlay_node[1] is not None:     # only update if object contains something (e.g. non empty list or dict) and is not None
                try:
                    # Try to call update() method
                    vars(self)[v_node[0]].update(v_overlay_node[1])
                except AttributeError:
                    # print("Object does not have update() method.")
                    vars(self)[v_node[0]] = v_overlay_node[1]
                except Exception as e:
                    print("Neither update() method could be called, nor simple value assignment was possible.")
                    print(e)
                    exit(-1)


class Node(Updateable):
    default_density = 200  # people/km^2
    default_population = 1000
    res_in_degrees = 2.5 / 60

    class IndividualProperty(Updateable):
        def __init__(self,
                     initial_distribution: List[float] = None,
                     property=None,
                     values: List[float] = None,
                     transitions: List[float] = None,
                     transmission_matrix: List[float] = None
                     ):
            self.initial_distribution = initial_distribution
            self.property = property
            self.values = values
            self.transitions = transitions
            self.transmission_matrix = transmission_matrix

        def to_dict(self) -> dict:
            individual_property = {
                "Initial_Distribution": self.initial_distribution,
                "Property": self.property,
                "Values": self.values,
                "Transitions": self.transitions
            }
            if self.transmission_matrix is not None:
                individual_property.update({"TransmissionMatrix": self.transmission_matrix})
            return individual_property

    class IndividualProperties(Updateable):
        def __init__(self, individual_property=None):
            self.individual_properties = [individual_property] if individual_property else None

        def add(self, individual_property):
            if not self.individual_properties:
                self.individual_properties = []
            self.individual_properties.append(individual_property)

        def to_dict(self) -> dict:
            individual_properties = []
            for ip in self.individual_properties:
                individual_properties.append(ip.to_dict())
            return individual_properties

        def __getitem__(self, index: int):
            return self.individual_properties[index]

        def __len__(self):
            if not self.individual_properties:
                return 0
            return len(self.individual_properties)

    class IndividualAttributes(Updateable):
        class SusceptibilityDistribution(Updateable):
            def __init__(self,
                         distribution_values: List[float] = None,
                         result_scale_factor=None,
                         result_values=None):
                self.distribution_values = distribution_values
                self.result_scale_factor = result_scale_factor
                self.result_values = result_values

            def to_dict(self) -> dict:
                susceptibility_distribution = {}

                if self.distribution_values is not None:
                    susceptibility_distribution.update({"DistributionValues": self.distribution_values})

                if self.result_scale_factor is not None:
                    susceptibility_distribution.update({"ResultScaleFactor": self.result_scale_factor})

                if self.result_values is not None:
                    susceptibility_distribution.update({"ResultValues": self.result_values})

                return susceptibility_distribution

        class AgeDistribution(Updateable):
            def __init__(self,
                         distribution_values=None,
                         result_scale_factor=None,
                         result_values=None):
                self.distribution_values = distribution_values
                self.result_scale_factor = result_scale_factor
                self.result_values = result_values

            def to_dict(self) -> dict:
                age_distribution = {}

                if self.distribution_values is not None:
                    age_distribution.update({"DistributionValues": self.distribution_values})

                if self.result_scale_factor is not None:
                    age_distribution.update({"ResultScaleFactor": self.result_scale_factor})

                if self.result_values is not None:
                    age_distribution.update({"ResultValues": self.result_values})

                return age_distribution

            def from_dict(self, age_distribution: dict):
                if age_distribution is not None:
                    self.distribution_values = age_distribution.get("DistributionValues")
                    self.result_scale_factor = age_distribution.get("ResultScaleFactor")
                    self.result_values = age_distribution.get("ResultValues")
                return self

        class MortalityDistribution(Updateable):
            def __init__(self,
                         axis_names: List[str] = None,
                         axis_scale_factor: List[float] = None,
                         axis_units=None,
                         num_distribution_axes=None,
                         num_population_axes=None,
                         num_population_groups=None,
                         population_groups=None,
                         result_scale_factor=None,
                         result_units=None,
                         result_values=None):
                self.axis_names = axis_names
                self.axis_scale_factor = axis_scale_factor
                self.axis_units = axis_units
                self.num_distribution_axes = num_distribution_axes
                self.num_population_axes = num_population_axes
                self.num_population_groups = num_population_groups
                self.population_groups = population_groups
                self.result_scale_factor = result_scale_factor
                self.result_units = result_units
                self.result_values = result_values

            def to_dict(self) -> dict:
                mortality_distribution = {}

                if self.axis_names is not None:
                    mortality_distribution.update({"AxisNames": self.axis_names})

                if self.axis_scale_factor is not None:
                    mortality_distribution.update({"AxisScaleFactors": self.axis_scale_factor})

                if self.axis_units is not None:
                    mortality_distribution.update({"AxisUnits": self.axis_units})

                if self.num_distribution_axes is not None:
                    mortality_distribution.update({"NumDistributionAxes": self.num_distribution_axes})

                if self.num_population_groups is not None:
                    mortality_distribution.update({"NumPopulationGroups": self.num_population_groups})

                if self.population_groups is not None:
                    mortality_distribution.update({"PopulationGroups": self.population_groups})

                if self.result_scale_factor is not None:
                    mortality_distribution.update({"ResultScaleFactor": self.result_scale_factor})

                if self.result_units is not None:
                    mortality_distribution.update({"ResultUnits": self.result_units})

                if self.result_values is not None:
                    mortality_distribution.update({"ResultValues": self.result_values})

                return mortality_distribution
        
            def from_dict(self, mortality_distribution: dict):
                if mortality_distribution:
                    self.axis_names = mortality_distribution.get("AxisNames")
                    self.axis_scale_factor = mortality_distribution.get("AxisScaleFactors")
                    self.axis_units = mortality_distribution.get("AxisUnits")
                    self.num_distribution_axes = mortality_distribution.get("NumDistributionAxes")
                    self.num_population_groups = mortality_distribution.get("NumPopulationGroups")
                    self.population_groups = mortality_distribution.get("PopulationGroups")
                    self.result_scale_factor = mortality_distribution.get("ResultScaleFactor")
                    self.result_units = mortality_distribution.get("ResultUnits")
                    self.result_values = mortality_distribution.get("ResultValues")
                return self

        def __init__(self,
                     age_distribution_flag=None,
                     age_distribution1=None,
                     age_distribution2=None,
                     age_distribution=None,
                     prevalence_distribution_flag=None,
                     prevalence_distribution1=None,
                     prevalence_distribution2=None,
                     immunity_distribution_flag=None,
                     immunity_distribution1=None,
                     immunity_distribution2=None,
                     risk_distribution_flag=None,
                     risk_distribution1=None,
                     risk_distribution2=None,
                     migration_heterogeneity_distribution_flag=None,
                     migration_heterogeneity_distribution1=None,
                     migration_heterogeneity_distribution2=None,
                     mortality_distribution=None,
                     susceptibility_distribution=None
                     ):
            self.age_distribution_flag = age_distribution_flag
            self.age_distribution1 = age_distribution1
            self.age_distribution2 = age_distribution2
            self.age_distribution = age_distribution
            self.prevalence_distribution_flag = prevalence_distribution_flag
            self.prevalence_distribution1 = prevalence_distribution1
            self.prevalence_distribution2 = prevalence_distribution2
            self.immunity_distribution_flag = immunity_distribution_flag
            self.immunity_distribution1 = immunity_distribution1
            self.immunity_distribution2 = immunity_distribution2
            self.risk_distribution_flag = risk_distribution_flag
            self.risk_distribution1 = risk_distribution1
            self.risk_distribution2 = risk_distribution2
            self.migration_heterogeneity_distribution_flag = migration_heterogeneity_distribution_flag
            self.migration_heterogeneity_distribution1 = migration_heterogeneity_distribution1
            self.migration_heterogeneity_distribution2 = migration_heterogeneity_distribution2
            self.mortality_distribution = mortality_distribution
            self.susceptibility_distribution = susceptibility_distribution

        def to_dict(self) -> dict:
            individual_attributes = {}
            if self.age_distribution_flag is not None:
                individual_attributes_age_distribution = {
                    "AgeDistributionFlag": self.age_distribution_flag,
                    "AgeDistribution1": self.age_distribution1,
                    "AgeDistribution2": self.age_distribution2
                }
                individual_attributes.update(individual_attributes_age_distribution)
            else:
                if self.age_distribution:
                    individual_attributes_age_distribution = {
                        "AgeDistribution": self.age_distribution.to_dict()
                    }
                    individual_attributes.update(individual_attributes_age_distribution)

            if self.prevalence_distribution_flag is not None:
                prevalence_distribution = {
                    "PrevalenceDistributionFlag": self.prevalence_distribution_flag,
                    "PrevalenceDistribution1": self.prevalence_distribution1,
                    "PrevalenceDistribution2": self.prevalence_distribution2
                }
                individual_attributes.update(prevalence_distribution)

            if self.immunity_distribution_flag is not None:
                immunity_distribution = {
                    "ImmunityDistributionFlag": self.immunity_distribution_flag,
                    "ImmunityDistribution1": self.immunity_distribution1,
                    "ImmunityDistribution2": self.immunity_distribution2
                }
                individual_attributes.update(immunity_distribution)

            if self.migration_heterogeneity_distribution_flag is not None:
                migration_heterogeneity_distribution = {
                    "MigrationHeterogeneityDistributionFlag": self.migration_heterogeneity_distribution_flag,
                    "MigrationHeterogeneityDistribution1": self.migration_heterogeneity_distribution1,
                    "MigrationHeterogeneityDistribution2": self.migration_heterogeneity_distribution2
                }
                individual_attributes.update(migration_heterogeneity_distribution)

            if self.risk_distribution_flag is not None:
                risk_distribution = {
                    "RiskDistributionFlag": self.risk_distribution_flag,
                    "RiskDistribution1": self.risk_distribution1,
                    "RiskDistribution2": self.risk_distribution2
                }
                individual_attributes.update(risk_distribution)

            if self.susceptibility_distribution is not None:
                individual_attributes.update({"SusceptibilityDistribution": self.susceptibility_distribution.to_dict()})

            if self.mortality_distribution is not None:
                individual_attributes.update({"MortalityDistribution": self.mortality_distribution.to_dict()})

            return individual_attributes

        def from_dict(self, individual_attributes: dict):
            self.age_distribution_flag = individual_attributes.get("AgeDistributionFlag")
            self.age_distribution1 = individual_attributes.get("AgeDistribution1")
            self.age_distribution2 = individual_attributes.get("AgeDistribution2")
            self.age_distribution = Node.IndividualAttributes.AgeDistribution().from_dict(individual_attributes.get("AgeDistribution"))  # DISTRIBUTION_COMPLEX
            self.prevalence_distribution_flag = individual_attributes.get("PrevalenceDistributionFlag")
            self.prevalence_distribution1 = individual_attributes.get("PrevalenceDistribution1")
            self.prevalence_distribution2 = individual_attributes.get("PrevalenceDistribution2")
            self.immunity_distribution_flag = individual_attributes.get("ImmunityDistributionFlag")
            self.immunity_distribution1 = individual_attributes.get("ImmunityDistribution1")
            self.immunity_distribution2 = individual_attributes.get("ImmunityDistribution2")
            self.risk_distribution_flag = individual_attributes.get("RiskDistributionFlag")
            self.risk_distribution1 = individual_attributes.get("RiskDistribution1")
            self.risk_distribution2 = individual_attributes.get("RiskDistribution2")
            self.migration_heterogeneity_distribution_flag = individual_attributes.get("MigrationHeterogeneityDistributionFlag")
            self.migration_heterogeneity_distribution1 = individual_attributes.get("MigrationHeterogeneityDistribution1")
            self.migration_heterogeneity_distribution2 = individual_attributes.get("MigrationHeterogeneityDistribution2")
            self.mortality_distribution = Node.IndividualAttributes.MortalityDistribution().from_dict(individual_attributes.get("MortalityDistribution"))
            return self

    class NodeAttributes(Updateable):
        def __init__(self,
                     airport: int = None,
                     altitude=None,
                     birth_rate: float = None,
                     country=None,
                     growth_rate: float = None,
                     name: str = None,
                     latitude: float = None,
                     longitude: float = None,
                     metadata: dict = None,
                     initial_population: int = None,
                     id: int = None,
                     region: int = None,
                     seaport: int = None,
                     larval_habitat_multiplier: List[float] = None,
                     initial_vectors_per_species=None
                     ):
            self.airport = airport
            self.altitude = altitude
            self.birth_rate = birth_rate
            self.country = country
            self.growth_rate = growth_rate
            self.Id = id
            self.initial_population = initial_population
            self.initial_vectors_per_species = initial_vectors_per_species
            self.larval_habitat_multiplier = larval_habitat_multiplier
            self.latitude = latitude
            self.longitude = longitude
            self.metadata = metadata
            self.name = name
            self.region = region
            self.seaport = seaport

        def from_dict(self, node_attributes: dict):
            self.airport = node_attributes.get("Airport")
            self.altitude = node_attributes.get("Altitude")
            self.country = node_attributes.get("country")
            self.growth_rate = node_attributes.get("GrowthRate")
            self.name = node_attributes.get("FacilityName")
            self.latitude = node_attributes.get("Latitude")
            self.longitude = node_attributes.get("Longitude")
            self.metadata = node_attributes.get("Metadata")
            self.initial_population = node_attributes.get("InitialPopulation")
            self.Id = node_attributes.get("Id")
            self.larval_habitat_multiplier = node_attributes.get("LarvalHabitatMultiplier")
            self.initial_vectors_per_species = node_attributes.get("InitialVectorsPerSpecies")
            self.birth_rate = node_attributes.get("BirthRate")
            self.seaport = node_attributes.get("Seaport")
            self.region = node_attributes.get("Region")
            return self

        def to_dict(self) -> dict:
            node_attributes = {}
            if self.birth_rate is not None:
                node_attributes.update({"BirthRate": self.birth_rate})

            if self.Id is not None:
                node_attributes.update({"Id": self.Id})

            if self.latitude is not None:
                node_attributes.update({"Latitude": self.latitude})

            if self.longitude is not None:
                node_attributes.update({"Longitude": self.longitude})

            if self.initial_population is not None:
                node_attributes.update({"InitialPopulation": int(self.initial_population)})

            if self.name:
                node_attributes.update({"FacilityName": self.name})

            if self.larval_habitat_multiplier is not None:
                node_attributes.update({"LarvalHabitatMultiplier": self.larval_habitat_multiplier})

            if self.initial_vectors_per_species:
                node_attributes.update({"InitialVectorsPerSpecies": self.initial_vectors_per_species})

            if self.airport is not None:
                node_attributes.update({"Airport": self.airport})

            if self.altitude is not None:
                node_attributes.update({"Altitude": self.altitude})

            if self.seaport is not None:
                node_attributes.update({"Seaport": self.seaport})

            if self.region is not None:
                node_attributes.update({"Region": self.region})

            if self.country is not None:
                node_attributes.update({"country": self.country})

            if self.growth_rate is not None:
                node_attributes.update({"GrowthRate": self.growth_rate})

            if self.metadata is not None:
                node_attributes.update({"Metadata": self.metadata})

            return node_attributes

    def __init__(
            self,
            lat,
            lon,
            pop,
            name: str = None,
            area: int = None,
            forced_id: int = None,
            extra_attributes: dict = None,
            individual_attributes: IndividualAttributes = None,
            individual_properties: IndividualProperties = None,
            node_attributes: NodeAttributes = None,
            meta: dict = None
    ):
        """
        Represent a Node
        :param lat: Latitude
        :param lon: Longitude
        :param pop: Population
        :param name:  Facility name
        :param area:  Area
        :param forced_id: Do we want a custom id instead of the normal ID based on lat/lon?
        :param extra_attributes: Extra node attributes
        """
        self.name = name
        self.forced_id = forced_id
        self.extra_attributes = extra_attributes if extra_attributes else {}
        self.meta = meta if meta else {}
        self.individual_attributes = individual_attributes if individual_attributes else Node.IndividualAttributes()
        self.individual_properties = individual_properties if individual_properties else Node.IndividualProperties()
        self.node_attributes = Node.NodeAttributes(latitude=lat, longitude=lon, initial_population=pop, name=name)
        if node_attributes is not None:
            self.node_attributes.update(node_attributes)
        if area:
            self.density = pop / area
        else:
            self.density = self.default_density

    def __repr__(self):
        return f"{self.node_attributes.name} - ({self.node_attributes.latitude},{self.node_attributes.longitude})"

    def to_dict(self) -> dict:
        if self.name:
            self.node_attributes.name = self.name

        if self.node_attributes:
            self.node_attributes.Id = self.forced_id

        d = {"NodeID": self.id,
             "NodeAttributes": self.node_attributes.to_dict()}

        if self.individual_attributes:
            d["IndividualAttributes"] = self.individual_attributes.to_dict()

        if self.individual_properties:
            ip_dict = {"IndividualProperties": []}
            for ip in self.individual_properties:
                ip_dict["IndividualProperties"].append(ip.to_dict())
            d.update(ip_dict)
        return d

    def to_tuple(self):
        return self.node_attributes.latitude, self.node_attributes.longitude, self.node_attributes.initial_population

    @property
    def id(self):
        return (
            self.forced_id
            if self.forced_id is not None
            else nodeid_from_lat_lon(self.node_attributes.latitude, self.node_attributes.longitude, self.res_in_degrees)
        )

    @classmethod
    def init_resolution_from_file(cls, fn):
        if "30arcsec" in fn:
            cls.res_in_degrees = 30 / 3600.0
        elif "2_5arcmin" in fn:
            cls.res_in_degrees = 2.5 / 30
        else:
            raise Exception("Don't recognize resolution from demographics filename")

    @classmethod
    def from_data(cls, data: dict):
        """
        Function used to create the node object from data (most likely coming from a demographics file)
        :param data: 
        :return: 
        """
        nodeid = data["NodeID"]
        node_attributes_dict = data.get("NodeAttributes")
        node_attributes_dict = dict(node_attributes_dict)
        attributes = data["NodeAttributes"]
        latitude = attributes.pop("Latitude")
        longitude = attributes.pop("Longitude")
        population = attributes.pop("InitialPopulation", Node.default_population)
        name = attributes.pop("FacilityName", nodeid)
        individual_attributes_dict = data.get("IndividualAttributes")
        individual_properties_dict = data.get("IndividualProperties")

        individual_properties = Node.IndividualProperties()
        if individual_properties_dict:
            if type(individual_properties_dict) is dict:
                individual_properties.append(individual_properties_dict)
            if type(individual_properties_dict) is list:
                for ip in individual_properties_dict:
                    individual_properties.add(Node.IndividualProperty(property=ip["Property"],
                                                                           values=ip["Values"],
                                                                           transitions=ip["Transitions"],
                                                                           initial_distribution=ip["Initial_Distribution"]))
        individual_attributes = None
        if individual_attributes_dict:
            individual_attributes = Node.IndividualAttributes().from_dict(individual_attributes_dict)

        node_attributes = None
        if node_attributes_dict:
            node_attributes = Node.NodeAttributes().from_dict(node_attributes_dict)

        # Create the node
        cls.node = Node(node_attributes.latitude, node_attributes.longitude, node_attributes.initial_population,
                        name=name, extra_attributes=attributes,
                        forced_id=nodeid, individual_attributes=individual_attributes,
                        individual_properties=individual_properties, node_attributes=node_attributes)
        return cls.node

    def set_individual_attributes(self, ind_attribute: IndividualAttributes):
        self.individual_attributes = ind_attribute

    def set_individual_properties(self, ind_properties: IndividualProperties):
        self.individual_properties = ind_properties

    def set_node_attributes(self, node_attributes: NodeAttributes):
        self.node_attributes = node_attributes


class OverlayNode(Node):
    """
    Node that only requires an ID. Use to overlay a Node.
    """

    def __init__(self,
                 node_id,
                 latitude=None,
                 longitude=None,
                 initial_population=None,
                 **kwargs
                 ):
        super(OverlayNode, self).__init__(latitude, longitude, initial_population,
                                          forced_id=node_id,
                                          **kwargs
                                          )


def get_xpix_ypix(nodeid):
    ypix = (nodeid - 1) & 2 ** 16 - 1
    xpix = (nodeid - 1) >> 16
    return (xpix, ypix)


def lat_lon_from_nodeid(nodeid, res_in_deg=Node.res_in_degrees):
    xpix, ypix = get_xpix_ypix(nodeid)
    lat = (0.5 + ypix) * res_in_deg - 90.0
    lon = (0.5 + xpix) * res_in_deg - 180.0
    return (lat, lon)


def xpix_ypix_from_lat_lon(lat, lon, res_in_deg=Node.res_in_degrees):
    xpix = int(math.floor((lon + 180.0) / res_in_deg))
    ypix = int(math.floor((lat + 90.0) / res_in_deg))
    return xpix, ypix


def nodeid_from_lat_lon(lat, lon, res_in_deg=Node.res_in_degrees):
    xpix, ypix = xpix_ypix_from_lat_lon(lat, lon, res_in_deg)
    nodeid = (xpix << 16) + ypix + 1
    return nodeid


def nodes_for_DTK(filename, nodes):
    with open(filename, "w") as f:
        json.dump(
            {"Nodes": [{"NodeID": n.id, "NodeAttributes": n.to_dict()} for n in nodes]},
            f,
            indent=4,
        )


def basicNode(lat: float = 0, lon: float = 0, pop: int = int(1e6), name: str = "node_name", forced_id: int = 1):
    return Node(lat, lon, pop, name=name, forced_id=forced_id)
